/*
** Module   : DataTypeHelper.java
** Abstract : helper to manage data types of objects referenced by HQL
**
** Copyright (c) 2004-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- --------------------------Description-------------------------------
** 001 ECF 20080310   @37492 Created initial version. Provides static
**                           helper methods to determine the data types of
**                           various constructs found within HQL.
** 002 ECF 20080609   @38681 Fixed initBasicTokens(). Added STRING/TYPE
**                           constant mapping to tokens map.
** 003 ECF 20080804   @39310 Added rowid support. Map Long and rowid
**                           classes, and long primitive, to HQLTypes.LONG
**                           enum.
** 004 SVL 20090225   @41366 Added recid support (mapped to
**                           HQLTypes.INTEGER).
** 005 GES 20090518   @42391 Import change.
** 006 ECF 20090817   @43683 Added persistence *Field types to classes map.
** 007 OM  20130501          Added support for 64 bit data types.
*/

package com.goldencode.p2j.persist.hql;

import java.lang.reflect.*;
import java.math.*;
import java.sql.Date;
import java.util.*;
import org.hibernate.type.Type;
import com.goldencode.ast.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.persist.pl.*;
import com.goldencode.p2j.persist.type.*;
import com.goldencode.p2j.util.*;

/**
 * A helper to manage data types of objects referenced by HQL statements
 * (specifically, where clauses).  This class provides static helper methods
 * to determine the data type of:
 * <ul>
 *   <li>certain non-volatile tokens (i.e., constants or other token types
 *       whose data types cannot change based on other factors)
 *   <li>standard functions, such as <code>upper()</code>
 *   <li>user defined functions, such as all the built-in functions migrated
 *       from Progress
 *   <li>query substitution parameters
 *   <li>qualified and unqualified references to DMO properties
 * </ul>
 * <p>
 * The {@link HQLTypes} enum is used to rationalize all the various type
 * constructs (token types, Java class types, Hibernate types) into a common
 * construct.  Any request for type information will be in terms of an
 * <code>HQLTypes</code> enum value.
 */
public final class DataTypeHelper
implements HQLParserTokenTypes
{
   /** Map of token types to HQL types */
   private static final Map<Integer, HQLTypes> tokens =
      new HashMap<Integer, HQLTypes>();
   
   /** Map of functions to HQL types */
   private static final Map<FunctionKey, HQLTypes> functions =
      new HashMap<FunctionKey, HQLTypes>();
   
   /** Map of query substitution parameter Hibernate types to HQL types */
   private static final Map<Type, HQLTypes> parameters =
      new HashMap<Type, HQLTypes>();
   
   /** Map of Java class types to HQL types */
   private static final Map<Class<?>, HQLTypes> classes =
      new HashMap<Class<?>, HQLTypes>();
   
   /**
    * Private constructor;  access is via static methods.
    */
   private DataTypeHelper()
   {
   }
   
   /**
    * Initialize the data structures needed to enable this helper.  This
    * should be invoked early during initialization of the persistence
    * runtime.
    */
   public static void initialize()
   {
      // Initialize Java data type wrapper mappings.
      initJavaClasses();
      
      // Initialize basic token type mappings.
      initBasicTokens();
      
      // Initialize query parameter type mappings (Hibernate types).
      initQueryParamTypes();
      
      // Initialize standard built-in function return type mappings.
      initFunctions();
   }
   
   /**
    * Get the HQL type associated with a particular Java class.
    * 
    * @param   javaType
    *          A Java class.
    * 
    * @return  The HQL type associated with the given class, or {@link
    *          HQLTypes#UNKNOWN} if no mapping is available.
    */
   public static HQLTypes getTypeClass(Class<?> javaType)
   {
      HQLTypes type = classes.get(javaType);
      
      return (type != null ? type : HQLTypes.UNKNOWN);
   }
   
   /**
    * Get the HQL type associated with a particular token type.  Only those
    * token types whose data type cannot vary (i.e., constants and certain
    * operators) are handled by this method.
    * 
    * @param   tokenType
    *          A token type from {@link HQLParserTokenTypes}.
    * 
    * @return  The HQL type associated with the given token type, or {@link
    *          HQLTypes#UNKNOWN} if no mapping is available.
    */
   public static HQLTypes getTypeToken(int tokenType)
   {
      HQLTypes type = tokens.get(tokenType);
      
      return (type != null ? type : HQLTypes.UNKNOWN);
   }
   
   /**
    * Get the HQL type associated with a particular Hibernate
    * <code>Type</code>, used to identify the data type of a query
    * substitution parameter.
    * 
    * @param   paramType
    *          A Hibernate <code>Type</code>.
    * 
    * @return  The HQL type associated with the given Hibernate type, or
    *          {@link HQLTypes#UNKNOWN} if no mapping is available.
    */
   public static HQLTypes getTypeParameter(Type paramType)
   {
      HQLTypes type = parameters.get(paramType);
      
      return (type != null ? type : HQLTypes.UNKNOWN);
   }
   
   /**
    * Get the HQL type associated with a particular {@link FunctionKey}.
    * 
    * @param   key
    *          A <code>FunctionKey</code> which describes a standard function
    *          or a built-in function.
    * 
    * @return  The HQL type associated with the given function key, or {@link
    *          HQLTypes#UNKNOWN} if no mapping is available.
    */
   public static HQLTypes getTypeFunction(FunctionKey key)
   {
      HQLTypes type = functions.get(key);
      
      return (type != null ? type : HQLTypes.UNKNOWN);
   }
   
   /**
    * Get the HQL type associated with a particular, qualified DMO property,
    * as represented by the given AST.
    * 
    * @param   aliasAst
    *          An AST which represents a DMO alias, whose first (and only)
    *          child represents a property of that DMO.
    * 
    * @return  The HQL type associated with the given property, or {@link
    *          HQLTypes#UNKNOWN} if no mapping is available.
    * 
    * @throws  PersistenceException
    *          if the property does not exist, or if there is some error
    *          determining its data type.
    */
   public static HQLTypes getTypeQualifiedProperty(Aast aliasAst)
   throws PersistenceException
   {
      Aast propAst = aliasAst.getChildAt(0);
      String alias = aliasAst.getText();
      String property = propAst.getText();
      Class<?> javaType = DBUtils.getDMOPropertyType(alias, property);
      
      return getTypeClass(javaType);
   }
   
   /**
    * Get the HQL type associated with a particular, unqualified DMO property,
    * as represented by the given AST.
    * 
    * @param   propAst
    *          An AST which represents a property of a DMO.
    * @param   dmoClass
    *          The DMO type.
    * 
    * @return  The HQL type associated with the given property, or {@link
    *          HQLTypes#UNKNOWN} if no mapping is available.
    * 
    * @throws  PersistenceException
    *          if the property does not exist, or if there is some error
    *          determining its data type.
    */
   public static HQLTypes getTypeUnqualifiedProperty(Aast propAst, Class<?> dmoClass)
   throws PersistenceException
   {
      String property = propAst.getText();
      Class<?> javaType = DBUtils.getDMOPropertyType(dmoClass, property);
      
      return getTypeClass(javaType);
   }
   
   /**
    * Determine the HQL type of the expression represented by the given AST
    * node, which may be a single leaf node, or the head of a more complex
    * expression.
    * 
    * @param   node
    *          AST node whose HQL type is to be determined.
    * @param   dmoClass
    *          Optional parameter which represents the type of DMO represented
    *          by the statement from which this expression is a part.  It may
    *          by <code>null</code> if all DMO property references within the
    *          expression are qualified by a DMO alias.  Otherwise, it must be
    *          non-<code>null</code>.
    * 
    * @return  The HQL type which represents the data type to which the given
    *          expression would evaluate.
    * 
    * @throws  PersistenceException
    *          if a DMO property referenced within the expression does not
    *          exist, or if there is some error determining its data type.
    */
   public static HQLTypes expressionType(Aast node, Class<?> dmoClass)
   throws PersistenceException
   {
      HQLTypes c0;
      HQLTypes c1;
      
      int tokType = node.getType();
      switch (tokType)
      {
         case ALIAS:
            return getTypeQualifiedProperty(node);
            
         case PROPERTY:
            return getTypeUnqualifiedProperty(node, dmoClass);
            
         case FUNCTION:
            return getTypeFunction(node, dmoClass);
            
         case SUBST:
            return HQLTypes.valueOf((String) node.getAnnotation("datatype"));
            
         case TERNARY:
            c0 = expressionType(node.getChildAt(1), dmoClass); // case...then
            c1 = expressionType(node.getChildAt(2), dmoClass); // case...else
            return (c0 == HQLTypes.DECIMAL || c1 == HQLTypes.DECIMAL
                    ? HQLTypes.DECIMAL
                    : c0);
            
         case PLUS:
         case MINUS:
         case MULTIPLY:
         case DIVIDE:
            c0 = expressionType(node.getChildAt(0), dmoClass);  // lvalue
            c1 = expressionType(node.getChildAt(1), dmoClass);  // rvalue
            return (c0 == HQLTypes.DECIMAL || c1 == HQLTypes.DECIMAL
                    ? HQLTypes.DECIMAL
                    : HQLTypes.INTEGER);
            
         case UN_MINUS:
            return expressionType(node.getChildAt(0), dmoClass);
            
         default:
            return getTypeToken(tokType);
      }
   }
   
   /**
    * Get the HQL type which will be returned when the given function is
    * executed.
    * 
    * @param   function
    *          AST node representing a standard or built-in function
    *          expression.
    * @param   dmoClass
    *          Optional parameter which represents the type of DMO represented
    *          by the statement from which this expression is a part.  It may
    *          by <code>null</code> if all DMO property references within the
    *          expression are qualified by a DMO alias.  Otherwise, it must be
    *          non-<code>null</code>.
    * 
    * @return  The HQL type associated with the given function, or {@link
    *          HQLTypes#UNKNOWN} if no mapping is available.
    * 
    * @throws  PersistenceException
    *          if a DMO property referenced within the expression does not
    *          exist, or if there is some error determining its data type.
    */
   private static HQLTypes getTypeFunction(Aast function, Class<?> dmoClass)
   throws PersistenceException
   {
      String name = function.getText();
      int len = function.getNumImmediateChildren();
      HQLTypes[] hqlSig = new HQLTypes[len];
      for (int i = 0; i < len; i++)
      {
         hqlSig[i] = expressionType(function.getChildAt(i), dmoClass);
      }
      FunctionKey key = new FunctionKey(name, hqlSig);
      
      return getTypeFunction(key);
   }
   
   /**
    * Initialize the mapping of Java classes to {@link HQLTypes} for those
    * classes which are known to be used in HQL where clauses.
    */
   private static void initJavaClasses()
   {
      // Java primitive types.
      classes.put(Boolean.TYPE     , HQLTypes.BOOLEAN);
      classes.put(Integer.TYPE     , HQLTypes.INTEGER);
      classes.put(Long.TYPE        , HQLTypes.LONG);
      classes.put(Double.TYPE      , HQLTypes.DOUBLE);
      
      // Standard Java wrapper types.
      classes.put(Boolean.class    , HQLTypes.BOOLEAN);
      classes.put(Integer.class    , HQLTypes.INTEGER);
      classes.put(Long.class       , HQLTypes.LONG);
      classes.put(BigDecimal.class , HQLTypes.DECIMAL);
      classes.put(Date.class       , HQLTypes.DATE);
      classes.put(String.class     , HQLTypes.TEXT);
      
      // P2J wrapper types.
      classes.put(logical.class    , HQLTypes.BOOLEAN);
      classes.put(int64.class      , HQLTypes.LONG);
      classes.put(integer.class    , HQLTypes.INTEGER);
      classes.put(recid.class      , HQLTypes.INTEGER);
      classes.put(decimal.class    , HQLTypes.DECIMAL);
      classes.put(date.class       , HQLTypes.DATE);
      classes.put(raw.class        , HQLTypes.RAW);
      classes.put(character.class  , HQLTypes.TEXT);
      classes.put(rowid.class      , HQLTypes.LONG);
      
      // Specialized P2J wrapper types.
      classes.put(LogicalField.class   , HQLTypes.BOOLEAN);
      classes.put(Int64Field.class     , HQLTypes.LONG);
      classes.put(IntegerField.class   , HQLTypes.INTEGER);
      classes.put(DecimalField.class   , HQLTypes.DECIMAL);
      classes.put(DateField.class      , HQLTypes.DATE);
      classes.put(RawField.class       , HQLTypes.RAW);
      classes.put(CharacterField.class , HQLTypes.TEXT);
   }
   
   /**
    * Initialize the mapping of basic token types to {@link HQLTypes}.  Only
    * those token types whose data types will not change (i.e., constants and
    * certain operators) are mapped here.
    */
   private static void initBasicTokens()
   {
      tokens.put(BOOL_TRUE  , HQLTypes.BOOLEAN);
      tokens.put(BOOL_FALSE , HQLTypes.BOOLEAN);
      tokens.put(IS_NULL    , HQLTypes.BOOLEAN);
      tokens.put(NOT_NULL   , HQLTypes.BOOLEAN);
      tokens.put(EQUALS     , HQLTypes.BOOLEAN);
      tokens.put(NOT_EQ     , HQLTypes.BOOLEAN);
      tokens.put(GT         , HQLTypes.BOOLEAN);
      tokens.put(LT         , HQLTypes.BOOLEAN);
      tokens.put(LTE        , HQLTypes.BOOLEAN);
      tokens.put(GTE        , HQLTypes.BOOLEAN);
      tokens.put(AND        , HQLTypes.BOOLEAN);
      tokens.put(OR         , HQLTypes.BOOLEAN);
      tokens.put(NOT        , HQLTypes.BOOLEAN);
      tokens.put(NUM_LITERAL, HQLTypes.INTEGER);
      tokens.put(DEC_LITERAL, HQLTypes.DECIMAL);
      tokens.put(STRING     , HQLTypes.TEXT);
   }
   
   /**
    * Initialize the mapping of Hibernate <code>Type</code>s to {@link
    * HQLTypes} for those Hibernate <code>Type</code>s (including custom P2J
    * wrapper types) which are known to be used in HQL where clauses as query
    * substitution parameters.
    */
   private static void initQueryParamTypes()
   {
      Map<Class<?>, Type> class2Type = DBUtils.getHibernateTypeMap();
      Iterator<Map.Entry<Class<?>, Type>> iter =
         class2Type.entrySet().iterator();
      while (iter.hasNext())
      {
         Map.Entry<Class<?>, Type> entry = iter.next();
         Class<?> clazz = entry.getKey();
         Type paramType = entry.getValue();
         HQLTypes hqlType = classes.get(clazz);
         
         if (hqlType == null)
         {
            throw new IllegalArgumentException(
               "No HQL type found for Java type " + clazz.getName());
         }
         
         parameters.put(paramType, hqlType);
      }
   }
   
   /**
    * Initialize the mappings of {@link FunctionKey} to {@link HQLTypes} for
    * all standard functions used in HQL (including built-in functions which
    * emulate Progress functionality).
    */
   private static void initFunctions()
   {
      initFunction("upper", HQLTypes.TEXT, HQLTypes.TEXT);
      
      Iterator<Method> iter = BuiltIns.getStandardFunctions().iterator();
      while (iter.hasNext())
      {
         initFunction(iter.next());
      }
   }
   
   /**
    * Given a Java method which represents a user-defined function accessible
    * from HQL, create a mapping of {@link FunctionKey} to {@link HQLTypes}
    * for that function.
    * 
    * @param   method
    *          Java method to be mapped.
    */
   private static void initFunction(Method method)
   {
      String name = method.getName();
      Class<?> returns = method.getReturnType();
      HQLTypes hqlReturns = classes.get(returns);
      
      if (hqlReturns == null)
      {
         throw new IllegalArgumentException(
            "No HQL type found for user defined function return type " +
            returns.getName());
      }
      
      Class<?>[] sig = method.getParameterTypes();
      HQLTypes[] hqlSig = new HQLTypes[sig.length];
      for (int i = 0; i < sig.length; i++)
      {
         Class<?> next = sig[i];
         HQLTypes hqlParam = classes.get(next);
         if (hqlParam == null)
         {
            throw new IllegalArgumentException(
               "No HQL type found for user defined function parameter type " +
               next.getName());
         }
         hqlSig[i] = hqlParam;
      }
      
      FunctionKey key = new FunctionKey(name, hqlSig);
      functions.put(key, hqlReturns);
   }
   
   /**
    * Given the name and signature of a user-defined function accessible from
    * HQL, create a mapping of {@link FunctionKey} to {@link HQLTypes} for
    * that function.
    * 
    * @param   name
    *          Function name
    * @param   returns
    *          Data type that the function returns.
    * @param   paramTypes
    *          Data types of function parameters, if any.
    */
   private static void initFunction(String name,
                                    HQLTypes returns,
                                    HQLTypes... paramTypes)
   {
      FunctionKey key = new FunctionKey(name, paramTypes);
      functions.put(key, returns);
   }
}
